
# This file is part of PyEMMA.
#
# Copyright (c) 2014-2017 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pickle import Pickler, Unpickler, UnpicklingError

import numpy as np
import logging
logger = logging.getLogger(__name__)

__author__ = 'marscher'


def _blosc_opts(complevel=9, complib='blosc:lz4', shuffle=True):
    shuffle = 2 if shuffle == 'bit' else 1 if shuffle else 0
    compressors = ['blosclz', 'lz4', 'lz4hc', 'snappy', 'zlib', 'zstd']
    complib = ['blosc:' + c for c in compressors].index(complib)
    args = {
        'compression': 32001,
        'compression_opts': (0, 0, 0, 0, complevel, shuffle, complib)
    }
    if shuffle:
        args['shuffle'] = False
    return args


def _check_blosc_avail():
    import tempfile, h5py
    blosc_opts = _blosc_opts()
    fid, name = tempfile.mkstemp()
    try:
        with h5py.File(name) as h5f:
            try:
                h5f.create_dataset('test', shape=(1,1), **blosc_opts)
            except ValueError as ve:
                if 'Unknown compression filter' in str(ve):
                    import warnings
                    warnings.warn('BLOSC compression filter unavailable. '
                                  'Your resulting file may be large and not optimal to process.')
                    return {}
                else:  # unknown exception
                    raise
            else:
                return blosc_opts
    finally:
        try:
            import os
            os.unlink(name)
        except:
            pass


# we cache this during runtime
_DEFAULT_BLOSC_OPTIONS = _check_blosc_avail()


class HDF5PersistentPickler(Pickler):
    # stores numpy arrays during pickling in given hdf5 group.
    def __init__(self, group, file):
        super().__init__(file=file, protocol=4)
        self.group = group
        self._seen_ids = set()

    def dump(self, *args, **kwargs):
        # we temporarily patch mdtraj.Topology to save state to numpy array
        from unittest import mock
        from pyemma._base.serialization.mdtraj_helpers import getstate
        with mock.patch('mdtraj.Topology.__getstate__', getstate, create=True):
            super(HDF5PersistentPickler, self).dump(*args, **kwargs)

    def _store(self, array):
        id_ = id(array)
        key = str(id_)
        # this actually makes no sense to check it here, however it is needed,
        # since there seems to be some race condition in h5py...
        if key in self.group:
            assert id_ in self._seen_ids
            return id_
        self._seen_ids.add(id_)
        self.group.create_dataset(name=key, data=array,
                                  chunks=True, **_DEFAULT_BLOSC_OPTIONS)
        return id_

    def persistent_id(self, obj):
        if (isinstance(obj, np.ndarray) and obj.dtype != np.object_
                and id(obj) not in self._seen_ids):
            # do not store empty arrays in hdf (more overhead)
            if not len(obj):
                return None
            array_id = self._store(obj)
            return 'np_array', array_id

        return None


class HDF5PersistentUnpickler(Unpickler):
    __allowed_packages = ('builtin',
                          'pyemma',
                          'mdtraj',
                          'numpy',
                          'scipy',
                          'bhmm',
                          )

    def __init__(self, group, file):
        super().__init__(file=file)
        self.group = group

    def persistent_load(self, pid):
        # This method is invoked whenever a persistent ID is encountered.
        # Here, pid is the type and the dataset id.
        type_tag, key_id = pid
        if type_tag == "np_array":
            return self.group[str(key_id)][:]
        else:
            # Always raises an error if you cannot return the correct object.
            # Otherwise, the unpickler will think None is the object referenced
            # by the persistent ID.
            raise UnpicklingError("unsupported persistent object")

    def load(self, *args, **kwargs):
        # we temporarily patch mdtraj.Topology to load state from numpy array
        from unittest import mock
        from pyemma._base.serialization.mdtraj_helpers import setstate
        with mock.patch('mdtraj.Topology.__setstate__', setstate, create=True):
            return super(HDF5PersistentUnpickler, self).load(*args, **kwargs)

    @staticmethod
    def __check_allowed(module):
        # check if we are allowed to unpickle from these modules.
        i = module.find('.')
        if i > 0:
            package = module[:i]
        else:
            package = module
        if package not in HDF5PersistentUnpickler.__allowed_packages:
            raise UnpicklingError('{mod} not allowed to unpickle'.format(mod=module))

    def find_class(self, module, name):
        self.__check_allowed(module)
        from .util import class_rename_registry
        new_class = class_rename_registry.find_replacement_for_old('{}.{}'.format(module, name))
        if new_class:
            return new_class
        return super(HDF5PersistentUnpickler, self).find_class(module, name)




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


r"""

.. currentmodule:: pyemma.msm

User Functions
==============
For most users, the following high-level functions are sufficient to estimate msm models from data.
Expert users may want to construct Estimators or Models (see below) directly.

.. autosummary::
   :toctree: generated/

   markov_model
   timescales_msm
   its
   estimate_markov_model
   bayesian_markov_model
   tpt
   timescales_hmsm
   estimate_hidden_markov_model
   bayesian_hidden_markov_model

MSM classes
===========

**Estimators** to generate models from data. If you are not an expert user,
use the API functions above.

.. autosummary::
   :toctree: generated/

   ImpliedTimescales
   ChapmanKolmogorovValidator
   MaximumLikelihoodMSM
   BayesianMSM
   OOMReweightedMSM
   MaximumLikelihoodHMSM
   BayesianHMSM


**Models** of the kinetics or stationary properties of the data.
If you are not an expert user, use the API functions above.

.. autosummary::
   :toctree: generated/

   MSM
   SampledMSM
   HMSM
   SampledHMSM
   ReactiveFlux
   PCCA


MSM functions (low-level API)
=============================
Low-level functions for estimation and analysis of transition matrices and io have been moved to `MSMTools
<https://msmtools.readthedocs.io/>`_.


"""

######################################################
from msmtools.analysis.dense.pcca import PCCA

#####################################################
# Estimators and models
from .models import MSM, HMSM, SampledMSM, SampledHMSM, ReactiveFlux

from .estimators import MaximumLikelihoodMSM, BayesianMSM
from .estimators import MaximumLikelihoodHMSM, BayesianHMSM
from .estimators import AugmentedMarkovModel, OOMReweightedMSM
from .estimators import ImpliedTimescales
from .estimators import ChapmanKolmogorovValidator


# high-level api
from .api import *




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




import unittest
import os

import numpy as np

from pyemma.coordinates.data import DataInMemory
from pyemma.coordinates.data import MDFeaturizer
from pyemma.coordinates import api
import msmtools.generation as msmgen
import tempfile

import pkg_resources
from pyemma.util.files import TemporaryDirectory
import pyemma.coordinates as coor

class TestPipeline(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        path = pkg_resources.resource_filename(__name__, 'data') + os.path.sep
        cls.pdb_file = os.path.join(path, 'bpti_ca.pdb')
        cls.feat = MDFeaturizer(cls.pdb_file)
        cls.feat.add_all()
        cls.traj_files = [
            os.path.join(path, 'bpti_001-033.xtc'),
            os.path.join(path, 'bpti_067-100.xtc')
        ]

        # generate HMM with two gaussians
        p = np.array([[0.99, 0.01], [0.01, 0.99]])
        t = 10000
        means = [np.array([-1, 1]), np.array([1, -1])]
        widths = [np.array([0.3, 2]), np.array([0.3, 2])]
        # continuous trajectory
        x = np.zeros((t, 2))
        # hidden trajectory
        dtraj = msmgen.generate_traj(p, t)
        for t in range(t):
            s = dtraj[t]
            x[t, 0] = widths[s][0] * np.random.randn() + means[s][0]
            x[t, 1] = widths[s][1] * np.random.randn() + means[s][1]
        cls.generated_data = x
        cls.generated_lag = 10

    def test_is_parametrized(self):
        # construct pipeline with all possible transformers
        p = api.pipeline(
            [
                api.source(self.traj_files, top=self.pdb_file),
                api.tica(),
                api.pca(),
                api.cluster_kmeans(k=50),
                api.cluster_regspace(dmin=50),
                api.cluster_uniform_time(k=20)
            ], run=False
        )
        self.assertFalse(p._is_estimated(), "If run=false, the pipeline should not be parametrized.")
        p.parametrize()
        self.assertTrue(p._is_estimated(), "If parametrized was called, the pipeline should be parametrized.")

    def test_np_reader_in_pipeline(self):
        with TemporaryDirectory() as td:
            file_name = os.path.join(td, "test.npy")
            data = np.random.random((100, 3))
            np.save(file_name, data)
            reader = api.source(file_name)
            p = api.pipeline(reader, run=False, stride=2, chunksize=5)
            p.parametrize()

    def test_add_element(self):
        # start with empty pipeline without auto-parametrization
        p = api.pipeline([], run=False)
        # add some reader
        reader = api.source(self.traj_files, top=self.pdb_file)
        p.add_element(reader)
        p.parametrize()

        # get the result immediately
        out1 = reader.get_output()

        # add some kmeans
        kmeans = api.cluster_kmeans(k=15)
        p.add_element(kmeans)
        p.parametrize()
        # get the result immediately
        kmeans1 = kmeans.get_output()

        # get reader output again
        out2 = reader.get_output()
        p.add_element(api.cluster_kmeans(k=2))
        p.parametrize()

        # get kmeans output again
        kmeans2 = kmeans.get_output()
        # check if add_element changes the intermediate results
        np.testing.assert_array_equal(out1[0], out2[0])
        np.testing.assert_array_equal(out1[1], out2[1])
        np.testing.assert_array_equal(kmeans1[0], kmeans2[0])
        np.testing.assert_array_equal(kmeans1[1], kmeans2[1])

    def test_set_element(self):
        reader = api.source(self.traj_files, top=self.pdb_file)
        pca = api.pca()
        p = api.pipeline([reader, pca])
        self.assertTrue(p._is_estimated())
        pca_out = pca.get_output()
        tica = api.tica(lag=self.generated_lag)
        # replace pca with tica
        p.set_element(1, tica)
        self.assertFalse(p._is_estimated(), "After replacing an element, the pipeline should not be parametrized.")
        p.parametrize()
        tica_out = tica.get_output()
        # check if replacement actually happened
        self.assertFalse(np.array_equal(pca_out[0], tica_out[0]),
                         "The output should not be the same when the method got replaced.")

    @unittest.skip("Known to be broken")
    def test_replace_data_source(self):
        reader_xtc = api.source(self.traj_files, top=self.pdb_file)
        reader_gen = DataInMemory(data=self.generated_data)

        kmeans = api.cluster_kmeans(k=10)
        assert hasattr(kmeans, '_chunks')
        p = api.pipeline([reader_xtc, kmeans])
        out1 = kmeans.get_output()
        # replace source
        print(reader_gen)
        p.set_element(0, reader_gen)
        assert hasattr(kmeans, '_chunks')
        p.parametrize()
        out2 = kmeans.get_output()
        self.assertFalse(np.array_equal(out1, out2), "Data source changed, so should the resulting clusters.")

    def test_discretizer(self):
        reader_gen = DataInMemory(data=self.generated_data)
        # check if exception safe
        api.discretizer(reader_gen)._chain[-1].get_output()
        api.discretizer(reader_gen, transform=api.tica())._chain[-1].get_output()
        api.discretizer(reader_gen, cluster=api.cluster_uniform_time())._chain[-1].get_output()
        api.discretizer(reader_gen, transform=api.pca(), cluster=api.cluster_regspace(dmin=10))._chain[-1].get_output()

    def test_no_cluster(self):
        reader_xtc = api.source(self.traj_files, top=self.pdb_file)
        # only reader
        api.pipeline(reader_xtc)
        reader_xtc.get_output()
        # reader + pca / tica
        tica = api.tica()
        pca = api.pca()
        api.pipeline([reader_xtc, tica])._chain[-1].get_output()
        api.pipeline([reader_xtc, pca])._chain[-1].get_output()

    def test_no_transform(self):
        reader_xtc = api.source(self.traj_files, top=self.pdb_file)
        api.pipeline([reader_xtc, api.cluster_kmeans(k=10)])._chain[-1].get_output()
        api.pipeline([reader_xtc, api.cluster_regspace(dmin=10)])._chain[-1].get_output()
        api.pipeline([reader_xtc, api.cluster_uniform_time()])._chain[-1].get_output()

    def test_chunksize(self):
        reader_xtc = api.source(self.traj_files, top=self.pdb_file)
        chunksize = 1001
        chain = [reader_xtc, api.tica(), api.cluster_mini_batch_kmeans(batch_size=0.3, k=3)]
        p = api.pipeline(chain, chunksize=chunksize, run=False)
        assert p.chunksize == chunksize
        for e in p._chain:
            assert e.chunksize == chunksize

if __name__ == "__main__":
    unittest.main()



import os
import unittest
from glob import glob

import numpy as np

from pyemma.coordinates.data import DataInMemory
from pyemma.util.contexts import settings
from pyemma.util.files import TemporaryDirectory


class TestCoordinatesIterator(unittest.TestCase):

    def setUp(self):
        self.d = [np.random.random((100, 3)).astype(np.float32) for _ in range(3)]

    def test_current_trajindex(self):
        r = DataInMemory(self.d)
        expected_itraj = 0
        for itraj, X in r.iterator(chunk=0):
            assert itraj == expected_itraj
            expected_itraj += 1

        expected_itraj = -1
        it = r.iterator(chunk=16)
        for itraj, X in it:
            if it.pos == 0:
                expected_itraj += 1
            assert itraj == expected_itraj == it.current_trajindex

    def test_n_chunks(self):
        r = DataInMemory(self.d)

        it0 = r.iterator(chunk=0)
        assert it0.n_chunks == 3  # 3 trajs

        it1 = r.iterator(chunk=50)
        assert it1.n_chunks == 3 * 2  # 2 chunks per trajectory

        it2 = r.iterator(chunk=30)
        # 3 full chunks and 1 small chunk per trajectory
        assert it2.n_chunks == 3 * 4

        it3 = r.iterator(chunk=30)
        it3.skip = 10
        assert it3.n_chunks == 3 * 3  # 3 full chunks per traj

        it4 = r.iterator(chunk=30)
        it4.skip = 5
        # 3 full chunks and 1 chunk of 5 frames per trajectory
        assert it4.n_chunks == 3 * 4

        # test for lagged iterator
        for stride in range(1, 5):
            for lag in range(0, 18):
                it = r.iterator(
                    lag=lag, chunk=30, stride=stride, return_trajindex=False)
                chunks = sum(1 for _ in it)
                np.testing.assert_equal(it.n_chunks, chunks,
                                        err_msg="Expected number of chunks did not agree with what the iterator "
                                                "returned for stride=%s, lag=%s" % (stride, lag))

        dd = [np.random.random((100, 3)), np.random.random((120, 3)), np.random.random((120, 3))]
        rr = DataInMemory(dd)

        # test for lagged iterator
        for stride in range(1, 5):
            for lag in [x for x in range(0, 18)] + [50, 100]:
                it = rr.iterator(lag=lag, chunk=30, stride=stride, return_trajindex=False)
                chunks = sum(1 for _ in it)
                np.testing.assert_equal(it.n_chunks, chunks,
                                        err_msg="Expected number of chunks did not agree with what the iterator "
                                                "returned for stride=%s, lag=%s" % (stride, lag))
                assert chunks == it.n_chunks

    def _count_chunks(self, it):
        with it:
            it.reset()
            nchunks = sum(1 for _ in it)
        self.assertEqual(it.n_chunks, nchunks, msg="{it}".format(it=it))

    def test_n_chunks_ra(self):
        """ """
        r = DataInMemory(self.d)

        def gen_sorted_stride(n):
            frames = np.random.randint(0, 99, size=n)
            trajs = np.random.randint(0, 3, size=n)

            stride = np.sort(np.stack((trajs, frames)).T, axis=1)
            # sort by file and frame index
            sort_inds = np.lexsort((stride[:, 1], stride[:, 0]))
            return stride[sort_inds]

        strides = [gen_sorted_stride(np.random.randint(1, 99)) for _ in range(10)]
        lengths = [len(x) for x in strides]
        for chunk in range(0, 100):#max(lengths)):
            for stride in strides:
                it = r.iterator(chunk=chunk, stride=stride)
                self._count_chunks(it)

    def test_chunksize(self):
        r = DataInMemory(self.d)
        cs = np.arange(1, 17)
        i = 0
        it = r.iterator(chunk=cs[i])
        for itraj, X in it:
            if not it.last_chunk_in_traj:
                assert len(X) == it.chunksize
            else:
                assert len(X) <= it.chunksize
            i += 1
            i %= len(cs)
            it.chunksize = cs[i]
            assert it.chunksize == cs[i]

    def test_chunksize_max_memory(self):
        from pyemma.util.contexts import settings
        data = np.random.random((10000, 10))
        max_size = 1024
        with settings(default_chunksize=str(max_size)):
            r = DataInMemory(data)
            for itraj, x in r.iterator():
                self.assertLessEqual(x.nbytes, max_size)

    def test_last_chunk(self):
        r = DataInMemory(self.d)
        it = r.iterator(chunk=0)
        for itraj, X in it:
            assert it.last_chunk_in_traj
            if itraj == 2:
                assert it.last_chunk

    def test_stride(self):
        r = DataInMemory(self.d)
        stride = np.arange(1, 17)
        i = 0
        it = r.iterator(stride=stride[i], chunk=1)
        for _ in it:
            i += 1
            i %= len(stride)
            it.stride = stride[i]
            assert it.stride == stride[i]

    def test_return_trajindex(self):
        r = DataInMemory(self.d)
        it = r.iterator(chunk=0)
        it.return_traj_index = True
        assert it.return_traj_index is True
        for tup in it:
            assert len(tup) == 2
        it.reset()
        it.return_traj_index = False
        assert it.return_traj_index is False
        itraj = 0
        for tup in it:
            np.testing.assert_equal(tup, self.d[itraj])
            itraj += 1

        for tup in r.iterator(return_trajindex=True):
            assert len(tup) == 2
        itraj = 0
        for tup in r.iterator(return_trajindex=False):
            np.testing.assert_equal(tup, self.d[itraj])
            itraj += 1

    def test_pos(self):
        r = DataInMemory(self.d)
        r.chunksize = 17
        it = r.iterator()
        t = 0
        for itraj, X in it:
            assert t == it.pos
            t += len(X)
            if it.last_chunk_in_traj:
                t = 0

    def test_write_to_csv_propagate_filenames(self):
        from pyemma.coordinates import source, tica
        with TemporaryDirectory() as td:
            data = [np.random.random((20, 3))] * 3
            fns = [os.path.join(td, f)
                   for f in ('blah.npy', 'blub.npy', 'foo.npy')]
            for x, fn in zip(data, fns):
                np.save(fn, x)
            reader = source(fns)
            assert reader.filenames == fns
            tica_obj = tica(reader, lag=1, dim=2)
            tica_obj.write_to_csv(extension=".exotic", chunksize=3)
            res = sorted([os.path.abspath(x) for x in glob(td + os.path.sep + '*.exotic')])
            self.assertEqual(len(res), len(fns))
            desired_fns = sorted([s.replace('.npy', '.exotic') for s in fns])
            self.assertEqual(res, desired_fns)

            # compare written results
            expected = tica_obj.get_output()
            actual = source(list(s.replace('.npy', '.exotic') for s in fns)).get_output()
            assert len(actual) == len(fns)
            for a, e in zip(actual, expected):
                np.testing.assert_allclose(a, e)

    def test_write_h5(self):
        from pyemma.coordinates import tica
        dim = 10
        data = [np.random.random((np.random.randint(50, 150), dim)) for _ in range(4)]
        tica = tica(data, lag=1)
        import tempfile
        out = tempfile.mktemp()
        group = '/test'
        def perform(chunksize, stride):
            try:
                transformed_output = tica.get_output(chunk=chunksize, stride=stride)
                tica.write_to_hdf5(out, group=group, chunksize=chunksize, stride=stride)

                import h5py
                with h5py.File(out) as f:
                    assert len(f[group]) == len(data)
                    for (itraj, actual), desired in zip(f[group].items(), transformed_output):
                        np.testing.assert_equal(actual, desired, err_msg='failed for cs=%s, stride=%s'
                                                                         %(chunksize, stride))
            finally:
                os.remove(out)

        for cs in [0, 1, 3, 10, 42, 50]:
            for s in [1, 2, 3, 10]:
                perform(cs, s)

        # test overwrite
        try:
            tica.write_to_hdf5(out, group=group)
            with self.assertRaises(ValueError):
                tica.write_to_hdf5(out, group=group)

            os.remove(out)
            tica.write_to_hdf5(out)
            with self.assertRaises(ValueError) as ctx:
                tica.write_to_hdf5(out)
            assert 'Refusing to overwrite data' in ctx.exception.args[0]

            os.remove(out)
            tica.write_to_hdf5(out, group=group)
            tica.write_to_hdf5(out, group=group, overwrite=True)

            os.remove(out)
            import h5py
            with h5py.File(out) as f:
                f.create_group('empty').create_dataset('0000', shape=(1,1))
            with self.assertRaises(ValueError):
                tica.write_to_hdf5(out, group='empty')
            tica.write_to_hdf5(out, group='empty', overwrite=True)
        finally:
            os.remove(out)

    def test_invalid_data_in_input_nan(self):
        self.d[0][-1] = np.nan
        r = DataInMemory(self.d)
        it = r.iterator()
        from pyemma.coordinates.data._base.datasource import InvalidDataInStreamException
        with settings(coordinates_check_output=True):
            with self.assertRaises(InvalidDataInStreamException):
                for itraj, X in it:
                    pass

    def test_invalid_data_in_input_inf(self):
        self.d[1][-1] = np.inf
        r = DataInMemory(self.d, chunksize=5)
        it = r.iterator()
        from pyemma.coordinates.data._base.datasource import InvalidDataInStreamException
        with settings(coordinates_check_output=True):
            with self.assertRaises(InvalidDataInStreamException) as cm:
                for itraj, X in it:
                    pass

    def test_lagged_iterator(self):
        import pyemma.coordinates as coor
        from pyemma.coordinates.tests.util import create_traj, get_top

        trajectory_length = 4720
        lagtime = 1000
        n_trajs = 15

        top = get_top()
        trajs_data = [create_traj(top=top, length=trajectory_length) for _ in range(n_trajs)]
        trajs = [t[0] for t in trajs_data]
        xyzs = [t[1].reshape(-1, 9) for t in trajs_data]

        reader = coor.source(trajs, top=top, chunksize=5000)

        for chunk in [None, 0, trajectory_length, trajectory_length+1, trajectory_length+1000]:
            it = reader.iterator(lag=lagtime, chunk=chunk, return_trajindex=True)
            with it:
                for itraj, X, Y in it:
                    np.testing.assert_equal(X.shape, Y.shape)
                    np.testing.assert_equal(X.shape[0], trajectory_length - lagtime)
                    np.testing.assert_array_almost_equal(X, xyzs[itraj][:trajectory_length-lagtime])
                    np.testing.assert_array_almost_equal(Y, xyzs[itraj][lagtime:])

    def test_lagged_iterator_optimized(self):
        import pyemma.coordinates as coor
        from pyemma.coordinates.tests.util import create_traj, get_top
        from pyemma.coordinates.util.patches import iterload

        trajectory_length = 4720
        lagtime = 20
        n_trajs = 15
        stride = iterload.MAX_STRIDE_SWITCH_TO_RA + 1

        top = get_top()
        trajs_data = [create_traj(top=top, length=trajectory_length) for _ in range(n_trajs)]
        trajs = [t[0] for t in trajs_data]
        xyzs = [t[1].reshape(-1, 9)[::stride] for t in trajs_data]
        xyzs_lagged = [t[1].reshape(-1, 9)[lagtime::stride] for t in trajs_data]

        reader = coor.source(trajs, stride=stride, top=top, chunksize=5000)

        memory_cutoff = iterload.MEMORY_CUTOFF
        try:
            iterload.MEMORY_CUTOFF = 8
            it = reader.iterator(stride=stride, lag=lagtime, chunk=5000, return_trajindex=True)
            with it:
                curr_itraj = 0
                t = 0
                for itraj, X, Y in it:
                    if itraj != curr_itraj:
                        curr_itraj = itraj
                        t = 0
                    np.testing.assert_equal(X.shape, Y.shape)
                    l = len(X)
                    np.testing.assert_array_almost_equal(X, xyzs[itraj][t:t+l])
                    np.testing.assert_array_almost_equal(Y, xyzs_lagged[itraj][t:t+l])
                    t += l
        finally:
            iterload.MEMORY_CUTOFF = memory_cutoff


if __name__ == '__main__':
    unittest.main()




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


r"""Implement a MSM class that builds a Markov state models from
microstate trajectories, automatically computes important properties
and provides them for later access.

.. moduleauthor:: F. Noe <frank DOT noe AT fu-berlin DOT de>

"""


from pyemma._base.model import SampledModel as _SampledModel
from pyemma.msm.models.hmsm import HMSM as _HMSM
from pyemma.util.types import is_iterable


class SampledHMSM(_HMSM, _SampledModel):
    r""" Sampled Hidden Markov state model """

    __serialize_version = 0

    # TODO: maybe rename to parametrize in order to avoid confusion with set_params that has a different behavior?
    def set_model_params(self, samples=None, conf=0.95,
                         P=None, pobs=None, pi=None, reversible=None, dt_model='1 step', neig=None):
        """

        Parameters
        ----------
        samples : list of MSM objects
            sampled MSMs
        conf : float, optional, default=0.68
            Confidence interval. By default one-sigma (68.3%) is used. Use 95.4% for two sigma or 99.7% for three sigma.

        """
        # set model parameters of superclass
        _SampledModel.set_model_params(self, samples=samples, conf=conf)
        _HMSM.set_model_params(self, P=P, pobs=pobs, pi=pi, reversible=reversible, dt_model=dt_model, neig=neig)

    def submodel(self, states=None, obs=None):
        """Returns a HMM with restricted state space

        Parameters
        ----------
        states : None or int-array
            Hidden states to restrict the model to (if not None).
        obs : None, str or int-array
            Observed states to restrict the model to (if not None).

        Returns
        -------
        hmm : HMM
            The restricted HMM.

        """
        # get the reference HMM submodel
        ref = super(SampledHMSM, self).submodel(states=states, obs=obs)
        # get the sample submodels
        samples_sub = [sample.submodel(states=states, obs=obs) for sample in self.samples]
        # new model
        return SampledHMSM(samples_sub, ref=ref, conf=self.conf)




import argparse
parser = argparse.ArgumentParser()

parser.add_argument('--lag', default=500)
parser.add_argument('--dest', default='run.sh')
args = parser.parse_args()

template = """#!/bin/bash

#SBATCH --job-name=final_lag_{lag}
#SBATCH -D /home/marscher/NO_BACKUP/final_lag_{lag}
#SBATCH --output=logs/cache_feat_%A_%a.out
#SBATCH --time=1-0
#SBATCH --mem=250000M
#SBATCH --cpus-per-task=24
#SBATCH --partition=big
#SBATCH --array 0-167
#SBATCH --export=PYTHONUNBUFFERED=1
#SBATCH --mail-type=fail
#SBATCH --mail-user=m.scherer@fu-berlin.de

export OMP_NUM_THREADS=$SLURM_CPUS_ON_NODE
/home/marscher/miniconda3/envs/cov_/bin/python /home/marscher/feature_sel/analysis/calc_cov_single_feat.py \
   --output=/home/marscher/NO_BACKUP/final_lag_{lag} --lag={lag} $SLURM_ARRAY_TASK_ID
""".format(lag=args.lag)

with open(args.dest, 'w') as f:
    f.write(template)




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


r"""
.. currentmodule:: pyemma.coordinates

User API
========

**Trajectory input/output and featurization**

.. autosummary::
   :toctree: generated/

   featurizer
   load
   source
   combine_sources
   pipeline
   discretizer
   save_traj
   save_trajs

**Covariance estimation**

.. autosummary::
  :toctree: generated/

   covariance_lagged

**Coordinate and feature transformations**

.. autosummary::
   :toctree: generated/

   pca
   tica
   vamp

**Clustering Algorithms**

.. autosummary::
   :toctree: generated/

   cluster_kmeans
   cluster_mini_batch_kmeans
   cluster_regspace
   cluster_uniform_time
   assign_to_centers

Classes
=======
**Coordinate classes** encapsulating complex functionality. You don't need to
construct these classes yourself, as this is done by the user API functions above.
Find here a documentation how to extract features from them.

**I/O and Featurization**

.. autosummary::
   :toctree: generated/

   data.MDFeaturizer
   data.CustomFeature

**Transformation estimators**

.. autosummary::
   :toctree: generated/

   transform.PCA
   transform.TICA
   transform.VAMP

**Covariance estimation**

.. autosummary::
   :toctree: generated/

   estimation.covariance.LaggedCovariance

**Clustering algorithms**

.. autosummary::
   :toctree: generated/

   clustering.KmeansClustering
   clustering.MiniBatchKmeansClustering
   clustering.RegularSpaceClustering
   clustering.UniformTimeClustering

**Transformers**

.. autosummary::
   :toctree: generated/

   data._base.transformer.StreamingTransformer
   pipelines.Pipeline

**Discretization**

.. autosummary::
   :toctree: generated/

   clustering.AssignCenters


"""
from .api import *




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import numpy as np
from pyemma.util.annotators import deprecated


__author__ = 'Fabian Paul'
__all__ = ['histogram']


@deprecated("Please use pyemma.coordinates.histogram()")
def hist(transform, dimensions, nbins):
    return histogram(transform, dimensions, nbins)


def histogram(transform, dimensions, nbins):
    '''Computes the N-dimensional histogram of the transformed data.

    Parameters
    ----------
    transform : pyemma.coordinates.transfrom.Transformer object
        transform that provides the input data
    dimensions : tuple of indices
        indices of the dimensions you want to examine
    nbins : tuple of ints
        number of bins along each dimension

    Returns
    -------
    counts : (bins[0],bins[1],...) ndarray of ints
        counts compatible with pyplot.pcolormesh and pyplot.bar
    edges : list of (bins[i]) ndarrays
        bin edges compatible with pyplot.pcolormesh and pyplot.bar,
        see below.

    Examples
    --------

    >>> import matplotlib.pyplot as plt # doctest: +SKIP

    Only for ipython notebook
    >> %matplotlib inline  # doctest: +SKIP

    >>> counts, edges=histogram(transform, dimensions=(0,1), nbins=(20, 30)) # doctest: +SKIP
    >>> plt.pcolormesh(edges[0], edges[1], counts.T) # doctest: +SKIP

    >>> counts, edges=histogram(transform, dimensions=(1,), nbins=(50,)) # doctest: +SKIP
    >>> plt.bar(edges[0][:-1], counts, width=edges[0][1:]-edges[0][:-1]) # doctest: +SKIP
    '''
    maximum = np.ones(len(dimensions)) * (-np.inf)
    minimum = np.ones(len(dimensions)) * np.inf
    # compute min and max
    for _, chunk in transform:
        maximum = np.max(
            np.vstack((
                maximum,
                np.max(chunk[:, dimensions], axis=0))),
            axis=0)
        minimum = np.min(
            np.vstack((
                minimum,
                np.min(chunk[:, dimensions], axis=0))),
            axis=0)
    # define bins
    bins = [np.linspace(m, M, num=n)
            for m, M, n in zip(minimum, maximum, nbins)]
    res = np.zeros(np.array(nbins) - 1)
    # compute actual histogram
    for _, chunk in transform:
        part, _ = np.histogramdd(chunk[:, dimensions], bins=bins)
        res += part
    return res, bins




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


r"""Unit test for the tpt-function

.. moduleauthor:: F.Noe <frank  DOT noe AT fu-berlin DOT de> 
.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>

"""

import unittest
import numpy as np
from pyemma.util.numeric import assert_allclose

from pyemma.msm import markov_model, estimate_markov_model, tpt
import msmtools.analysis as msmana


class TestReactiveFluxFunctions(unittest.TestCase):
    def setUp(self):
        # 5-state toy system
        self.P = np.array([[0.8, 0.15, 0.05, 0.0, 0.0],
                           [0.1, 0.75, 0.05, 0.05, 0.05],
                           [0.05, 0.1, 0.8, 0.0, 0.05],
                           [0.0, 0.2, 0.0, 0.8, 0.0],
                           [0.0, 0.02, 0.02, 0.0, 0.96]])
        self.A = [0]
        self.B = [4]
        self.I = [1, 2, 3]

        # REFERENCE SOLUTION FOR PATH DECOMP
        self.ref_committor = np.array([0., 0.35714286, 0.42857143, 0.35714286, 1.])
        self.ref_backwardcommittor = np.array([1., 0.65384615, 0.53125, 0.65384615, 0.])
        self.ref_grossflux = np.array([[0., 0.00771792, 0.00308717, 0., 0.],
                                       [0., 0., 0.00308717, 0.00257264, 0.00720339],
                                       [0., 0.00257264, 0., 0., 0.00360169],
                                       [0., 0.00257264, 0., 0., 0.],
                                       [0., 0., 0., 0., 0.]])
        self.ref_netflux = np.array([[0.00000000e+00, 7.71791768e-03, 3.08716707e-03, 0.00000000e+00, 0.00000000e+00],
                                     [0.00000000e+00, 0.00000000e+00, 5.14527845e-04, 0.00000000e+00, 7.20338983e-03],
                                     [0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 3.60169492e-03],
                                     [0.00000000e+00, 4.33680869e-19, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00],
                                     [0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00]])
        self.ref_totalflux = 0.0108050847458
        self.ref_kAB = 0.0272727272727
        self.ref_mfptAB = 36.6666666667

        self.ref_paths = [[0, 1, 4], [0, 2, 4], [0, 1, 2, 4]]
        self.ref_pathfluxes = np.array([0.00720338983051, 0.00308716707022, 0.000514527845036])

        self.ref_paths_99percent = [[0, 1, 4], [0, 2, 4]]
        self.ref_pathfluxes_99percent = np.array([0.00720338983051, 0.00308716707022])
        self.ref_majorflux_99percent = np.array([[0., 0.00720339, 0.00308717, 0., 0.],
                                                 [0., 0., 0., 0., 0.00720339],
                                                 [0., 0., 0., 0., 0.00308717],
                                                 [0., 0., 0., 0., 0.],
                                                 [0., 0., 0., 0., 0.]])

        msmobj = markov_model(self.P)
        msmobj.mu = msmana.statdist(self.P)
        msmobj.estimated = True
        msmobj1 = msmobj

        # Testing:
        # self.tpt1 = tpt(self.P, self.A, self.B)
        self.tpt1 = tpt(msmobj1, self.A, self.B)

        # 16-state toy system
        P2_nonrev = np.array([[0.5, 0.2, 0.0, 0.0, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                              [0.2, 0.5, 0.1, 0.0, 0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                              [0.0, 0.1, 0.5, 0.2, 0.0, 0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                              [0.0, 0.0, 0.1, 0.5, 0.0, 0.0, 0.0, 0.4, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                              [0.3, 0.0, 0.0, 0.0, 0.5, 0.1, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                              [0.0, 0.1, 0.0, 0.0, 0.2, 0.5, 0.1, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                              [0.0, 0.0, 0.1, 0.0, 0.0, 0.1, 0.5, 0.2, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0],
                              [0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 0.3, 0.5, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0],
                              [0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.5, 0.1, 0.0, 0.0, 0.3, 0.0, 0.0, 0.0],
                              [0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 0.2, 0.5, 0.1, 0.0, 0.0, 0.1, 0.0, 0.0],
                              [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 0.1, 0.5, 0.1, 0.0, 0.0, 0.2, 0.0],
                              [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 0.2, 0.5, 0.0, 0.0, 0.0, 0.2],
                              [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.3, 0.0, 0.0, 0.0, 0.5, 0.2, 0.0, 0.0],
                              [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 0.3, 0.5, 0.1, 0.0],
                              [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.0, 0.1, 0.5, 0.2],
                              [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.3, 0.0, 0.0, 0.2, 0.5]])
        pstat2_nonrev = msmana.statdist(P2_nonrev)
        # make reversible
        C = np.dot(np.diag(pstat2_nonrev), P2_nonrev)
        Csym = C + C.T
        self.P2 = Csym / np.sum(Csym, axis=1)[:, np.newaxis]
        pstat2 = msmana.statdist(self.P2)
        self.A2 = [0, 4]
        self.B2 = [11, 15]
        self.coarsesets2 = [[2, 3, 6, 7], [10, 11, 14, 15], [0, 1, 4, 5], [8, 9, 12, 13], ]

        # REFERENCE SOLUTION CG
        self.ref2_tpt_sets = [set([0, 4]), set([2, 3, 6, 7]), set([10, 14]), set([1, 5]), set([8, 9, 12, 13]),
                              set([11, 15])]
        self.ref2_cgA = [0]
        self.ref2_cgI = [1, 2, 3, 4]
        self.ref2_cgB = [5]
        self.ref2_cgpstat = np.array([0.15995388, 0.18360442, 0.12990937, 0.11002342, 0.31928127, 0.09722765])
        self.ref2_cgcommittor = np.array([0., 0.56060272, 0.73052426, 0.19770537, 0.36514272, 1.])
        self.ref2_cgbackwardcommittor = np.array([1., 0.43939728, 0.26947574, 0.80229463, 0.63485728, 0.])
        self.ref2_cggrossflux = np.array([[0., 0., 0., 0.00427986, 0.00282259, 0.],
                                          [0., 0, 0.00234578, 0.00104307, 0., 0.00201899],
                                          [0., 0.00113892, 0, 0., 0.00142583, 0.00508346],
                                          [0., 0.00426892, 0., 0, 0.00190226, 0.],
                                          [0., 0., 0.00530243, 0.00084825, 0, 0.],
                                          [0., 0., 0., 0., 0., 0.]])
        self.ref2_cgnetflux = np.array([[0., 0., 0., 0.00427986, 0.00282259, 0.],
                                        [0., 0., 0.00120686, 0., 0., 0.00201899],
                                        [0., 0., 0., 0., 0., 0.00508346],
                                        [0., 0.00322585, 0., 0., 0.00105401, 0.],
                                        [0., 0., 0.0038766, 0., 0., 0.],
                                        [0., 0., 0., 0., 0., 0.]])

        """Dummy dtraj to trick trick constructor of MSM"""
        dtraj = [0, 0]
        tau = 1

        msmobj = markov_model(self.P2)
        msmobj.mu = msmana.statdist(self.P2)
        msmobj.estimated = True
        msmobj2 = msmobj

        # Testing
        self.tpt2 = tpt(msmobj2, self.A2, self.B2)

    def test_nstates(self):
        self.assertEqual(self.tpt1.nstates, np.shape(self.P)[0])

    def test_A(self):
        self.assertEqual(self.tpt1.A, self.A)

    def test_I(self):
        self.assertEqual(self.tpt1.I, self.I)

    def test_B(self):
        self.assertEqual(self.tpt1.B, self.B)

    def test_flux(self):
        assert_allclose(self.tpt1.flux, self.ref_netflux, rtol=1e-02, atol=1e-07)

    def test_netflux(self):
        assert_allclose(self.tpt1.net_flux, self.ref_netflux, rtol=1e-02, atol=1e-07)

    def test_grossflux(self):
        assert_allclose(self.tpt1.gross_flux, self.ref_grossflux, rtol=1e-02, atol=1e-07)

    def test_committor(self):
        assert_allclose(self.tpt1.committor, self.ref_committor, rtol=1e-02, atol=1e-07)

    def test_forwardcommittor(self):
        assert_allclose(self.tpt1.forward_committor, self.ref_committor, rtol=1e-02, atol=1e-07)

    def test_backwardcommittor(self):
        assert_allclose(self.tpt1.backward_committor, self.ref_backwardcommittor, rtol=1e-02, atol=1e-07)

    def test_total_flux(self):
        assert_allclose(self.tpt1.total_flux, self.ref_totalflux, rtol=1e-02, atol=1e-07)

    def test_rate(self):
        assert_allclose(self.tpt1.rate, self.ref_kAB, rtol=1e-02, atol=1e-07)
        assert_allclose(1.0 / self.tpt1.rate, self.ref_mfptAB, rtol=1e-02, atol=1e-07)

    """These tests are broken since the matrix is irreversible and no
    pathway-decomposition can be computed for irreversible matrices"""
    # def test_pathways(self):
    # # all paths
    #     (paths,pathfluxes) = self.tpt1.pathways()
    #     self.assertEqual(len(paths), len(self.ref_paths))
    #     for i in range(len(paths)):
    #         self.assertTrue(np.all(np.array(paths[i]) == np.array(self.ref_paths[i])))
    #     assert_allclose(pathfluxes, self.ref_pathfluxes, rtol=1e-02, atol=1e-07)
    #     # major paths
    #     (paths,pathfluxes) = self.tpt1.pathways(fraction = 0.99)
    #     self.assertEqual(len(paths), len(self.ref_paths_99percent))
    #     for i in range(len(paths)):
    #         self.assertTrue(np.all(np.array(paths[i]) == np.array(self.ref_paths_99percent[i])))
    #     assert_allclose(pathfluxes, self.ref_pathfluxes_99percent, rtol=1e-02, atol=1e-07)

    # def test_major_flux(self):
    #     # all flux
    #     assert_allclose(self.tpt1.major_flux(fraction=1.0), self.ref_netflux, rtol=1e-02, atol=1e-07)
    #     # 0.99 flux
    #     assert_allclose(self.tpt1.major_flux(fraction=0.99), self.ref_majorflux_99percent, rtol=1e-02, atol=1e-07)

    def test_coarse_grain(self):
        (tpt_sets, cgRF) = self.tpt2.coarse_grain(self.coarsesets2)
        self.assertEqual(tpt_sets, self.ref2_tpt_sets)
        self.assertEqual(cgRF.A, self.ref2_cgA)
        self.assertEqual(cgRF.I, self.ref2_cgI)
        self.assertEqual(cgRF.B, self.ref2_cgB)
        assert_allclose(cgRF.stationary_distribution, self.ref2_cgpstat)
        assert_allclose(cgRF.committor, self.ref2_cgcommittor)
        assert_allclose(cgRF.forward_committor, self.ref2_cgcommittor)
        assert_allclose(cgRF.backward_committor, self.ref2_cgbackwardcommittor)
        assert_allclose(cgRF.flux, self.ref2_cgnetflux)
        assert_allclose(cgRF.net_flux, self.ref2_cgnetflux)
        assert_allclose(cgRF.gross_flux, self.ref2_cggrossflux)

    def test_time_units(self):
        dtraj = np.random.randint(0, 4, 1000)
        tau = 12
        dt = 0.456
        msmobj = estimate_markov_model(dtraj, lag=tau, dt_traj='%f ns' % dt)

        # check MFPT consistency
        mfpt_ref = msmobj.mfpt([0], [1])
        tptobj = tpt(msmobj, [0], [1])
        assert_allclose(tptobj.mfpt, mfpt_ref)
        assert_allclose(msmana.mfpt(msmobj.P, [1], [0], tau=tau) * dt, mfpt_ref)
        assert_allclose(np.dot(msmobj.stationary_distribution, tptobj.backward_committor) / tptobj.total_flux, mfpt_ref)

        # check flux consistency
        total_flux_ref = tptobj.total_flux
        A = tptobj.A
        B = tptobj.B
        I = tptobj.I
        assert_allclose(tptobj.gross_flux[A, :][:, B].sum() + tptobj.gross_flux[A, :][:, I].sum(),
                        total_flux_ref)
        assert_allclose(tptobj.net_flux[A, :][:, B].sum() + tptobj.net_flux[A, :][:, I].sum(), total_flux_ref)
        assert_allclose(tptobj.flux[A, :][:, B].sum() + tptobj.flux[A, :][:, I].sum(), total_flux_ref)
        mf = tptobj.major_flux(1.0)
        assert_allclose(mf[A, :][:, B].sum() + mf[A, :][:, I].sum(), total_flux_ref)

        # check that the coarse-grained version is consistent too
        _, tptobj2 = tptobj.coarse_grain([A, I, B])
        assert_allclose(tptobj2.total_flux, total_flux_ref)
        assert_allclose(tptobj2.mfpt, mfpt_ref)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from pyemma.msm.estimators._dtraj_stats import DiscreteTrajectoryStats, blocksplit_dtrajs, cvsplit_dtrajs
from pyemma.util.types import ensure_dtraj_list
import msmtools


class TestDtrajStats(unittest.TestCase):

    def test_blocksplit_dtrajs_sliding(self):
        dtrajs = [np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]), np.array([0, 1, 9, 10])]
        for lag in range(1, 10):
            dtrajs_new = blocksplit_dtrajs(dtrajs, lag=lag, sliding=True)
            C1 = msmtools.estimation.count_matrix(dtrajs, lag, sliding=True, nstates=11).toarray()
            C2 = msmtools.estimation.count_matrix(dtrajs_new, lag, sliding=True, nstates=11).toarray()
            assert np.all(C1 == C2)

    def test_blocksplit_dtrajs_sampling(self):
        dtrajs = [np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]), np.array([0, 1, 9, 10])]
        for lag in range(1, 10):
            dtrajs_new = blocksplit_dtrajs(dtrajs, lag=lag, sliding=False, shift=0)
            C1 = msmtools.estimation.count_matrix(dtrajs, lag, sliding=False, nstates=11).toarray()
            C2 = msmtools.estimation.count_matrix(dtrajs_new, lag, sliding=False, nstates=11).toarray()
            assert np.all(C1 == C2)

    def test_blocksplit_dtrajs_cvsplit(self):
        dtrajs = [np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]), np.array([0, 1, 9, 10])]
        for lag in range(1, 5):
            dtrajs_new = blocksplit_dtrajs(dtrajs, lag=lag, sliding=False, shift=0)
            dtrajs_train, dtrajs_test = cvsplit_dtrajs(dtrajs_new)
            dtrajs_train = ensure_dtraj_list(dtrajs_train)
            dtrajs_test = ensure_dtraj_list(dtrajs_test)
            assert len(dtrajs_train) > 0
            assert len(dtrajs_test) > 0

    def test_mincount_connectivity(self):
        dtrajs = np.zeros(10, dtype=int)
        dtrajs[0] = 1
        dtrajs[-1] = 1
        dts = DiscreteTrajectoryStats(dtrajs)

        dts.count_lagged(1, mincount_connectivity=0)

        C_mincount0 = dts.count_matrix_largest.todense()

        np.testing.assert_equal(C_mincount0, np.array([[7, 1], [1, 0]]))

        dts.count_lagged(1, mincount_connectivity=2)
        C = np.array(dts.count_matrix_largest.todense())
        np.testing.assert_equal(C, np.array([[7]]))

        # check that the original count matrix remains unmodified
        np.testing.assert_equal(dts.count_matrix().todense(), C_mincount0)

    def test_core_sets(self):
        dtrajs   = [np.array([0, 0, 2, 0, 0, 3, 0, 5, 5, 5, 0, 0, 6, 8, 4, 1, 2, 0, 3])]
        expected = [np.array([-1, -1, 2, 2, 2, 3, 3, 5, 5, 5, 5, 5, 6, 6, 4, 4, 2, 2, 3])]
        core_set = np.arange(2, 7)
        dts = DiscreteTrajectoryStats(dtrajs)
        dts.to_coreset(core_set=core_set)

        np.testing.assert_equal(dts.discrete_trajectories, expected)

    def test_core_sets_2(self):
        dtrajs = [np.array([0, 0, 2, 1, 2])]
        expected = [np.array([-1, -1, 2, 1, 2])]
        dts = DiscreteTrajectoryStats(dtrajs)
        dts.to_coreset(np.arange(1, 3))
        np.testing.assert_equal(dts.discrete_trajectories,
                                expected)

    def test_core_sets_3(self):
        dtrajs = [np.array([2, 0, 1, 1, 2])]
        expected = [np.array([2, 2, 1, 1, 2])]
        dts = DiscreteTrajectoryStats(dtrajs)
        dts.to_coreset(np.arange(1, 3))
        np.testing.assert_equal(dts.discrete_trajectories,
                                expected)

    def test_core_sets_4(self):
        dtrajs = [np.array([2, 0, 0, 2, 0, 2, 0, 2])]
        dts = DiscreteTrajectoryStats(dtrajs)
        dts.to_coreset([2])
        np.testing.assert_equal(dts.discrete_trajectories,
                                [np.ones_like(dtrajs[0])*2])

    def test_core_sets_5(self):
        dtrajs = [np.array([2, 2, 2, 2, 2, 2, 2, 0])]
        dts = DiscreteTrajectoryStats(dtrajs)
        dts.to_coreset([2])
        np.testing.assert_equal(dts.discrete_trajectories,
                                [np.ones_like(dtrajs[0]) * 2])

    def test_realistic_random(self):
        n_states = 102
        n_traj = 10
        dtrajs = [np.random.randint(0, n_states, size=1000) for _ in range(n_traj)]
        core_set = np.random.randint(0, n_states, size=30)
        dts = DiscreteTrajectoryStats(dtrajs)
        actual = dts.to_coreset(core_set, in_place=False)

        def naive(dtrajs, core_set):
            import copy
            dtrajs = copy.deepcopy(dtrajs)
            newdiscretetraj = []
            for t, st in enumerate(dtrajs):
                oldmicro = -1
                newtraj = []
                for f, micro in enumerate(st):
                    newmicro = None
                    for co in core_set:
                        if micro == co:
                            newmicro = micro
                            oldmicro = micro
                            break
                    if newmicro is None and oldmicro is not None:
                        newtraj.append(oldmicro)
                    elif newmicro is not None:
                        newtraj.append(newmicro)
                    else:
                        print("hi there")
                        newtraj.append(-1)
                newdiscretetraj.append(np.array(newtraj, dtype=int))

            return newdiscretetraj

        expected = naive(dtrajs, core_set)
        np.testing.assert_equal(actual, expected)


if __name__ == '__main__':
    unittest.main()



# This file is part of PyEMMA.
#
# Copyright (c) 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
Created on 25.05.2016

@author: marscher
'''

import itertools
import os
import time
import warnings
from io import BytesIO
from operator import itemgetter

import numpy as np

from pyemma.coordinates.data.util.traj_info_cache import (UnknownDBFormatException,
                                                          TrajInfo,
                                                          TrajectoryInfoCache,
                                                          logger)
from pyemma.util import config


class AbstractDB(object):
    def set(self, value):
        # value: TrajInfo
        pass

    def update(self, value):
        pass

    def close(self):
        pass

    def sync(self):
        pass

    def get(self, key):
        # should raise KeyError in case of non existent key
        pass

    @property
    def db_version(self):
        pass

    @db_version.setter
    def db_version(self, val):
        pass


class DictDB(AbstractDB):
    def __init__(self):
        self._db = {}
        self.db_version = TrajectoryInfoCache.DB_VERSION

    def set(self, value):
        self._db[value.hash_value] = value

    def update(self, value):
        self._db[value.hash_value] = value

    @property
    def db_version(self):
        return self._db['version']

    @db_version.setter
    def db_version(self, version):
        self._db['version'] = version

    @property
    def num_entries(self):
        return len(self._db) - 1  # substract field for db_version


class SqliteDB(AbstractDB):
    def __init__(self, filename=None, clean_n_entries=30):
        """
        :param filename: path to database file
        :param clean_n_entries: during cleaning delete n % entries.
        """
        self.clean_n_entries = clean_n_entries
        import sqlite3

        # register numpy array conversion functions
        # uses "np_array" type in sql tables
        def adapt_array(arr):
            out = BytesIO()
            np.savez_compressed(out, offsets=arr)
            out.seek(0)
            return out.read()

        def convert_array(text):
            out = BytesIO(text)
            out.seek(0)
            npz = np.load(out)
            arr = npz['offsets']
            npz.close()
            return arr
        # Converts np.array to TEXT when inserting
        sqlite3.register_adapter(np.ndarray, adapt_array)

        # Converts TEXT to np.array when selecting
        sqlite3.register_converter("NPARRAY", convert_array)
        self._database = sqlite3.connect(filename if filename is not None else ":memory:",
                                         detect_types=sqlite3.PARSE_DECLTYPES, timeout=5,
                                         isolation_level=None)
        self.filename = filename

        self.lru_timeout = 5.0 # python sqlite3 specifies timeout in seconds instead of milliseconds.

        try:
            cursor = self._database.execute("select num from version")
            row = cursor.fetchone()
            if not row:
                self.db_version = TrajectoryInfoCache.DB_VERSION
                version = self.db_version
            else:
                version = row[0]
            if version != TrajectoryInfoCache.DB_VERSION:
                # drop old db? or try to convert?
                self._create_new_db()
        except sqlite3.OperationalError as e:
            if "no such table" in str(e):
                self._create_new_db()
                self.db_version = TrajectoryInfoCache.DB_VERSION
        except sqlite3.DatabaseError:
            bak = filename + ".bak"
            warnings.warn("TrajInfo database corrupted. Backing up file to %s and start with new one." % bak)
            self._database.close()
            import shutil
            shutil.move(filename, bak)
            SqliteDB.__init__(self, filename)

    def _create_new_db(self):
        # assumes self.database is a sqlite3.Connection
        create_version_table = "CREATE TABLE IF NOT EXISTS version (num INTEGER PRIMARY KEY);"
        create_info_table = """CREATE TABLE IF NOT EXISTS  traj_info(
            hash VARCHAR(64) PRIMARY KEY,
            length INTEGER,
            ndim INTEGER,
            offsets NPARRAY,
            abs_path VARCHAR(4096) null,
            version INTEGER,
            lru_db INTEGER
        );
        """
        self._database.execute(create_version_table)
        self._database.execute(create_info_table)
        self._database.commit()

    def close(self):
        self._database.close()

    @property
    def db_version(self):
        cursor = self._database.execute("select num from version")
        row = cursor.fetchone()
        if not row:
            raise RuntimeError("unknown db version")
        return row[0]

    @db_version.setter
    def db_version(self, val):
        import sqlite3
        with self._database:
            try:
                self._database.execute("insert into version VALUES (?)", [val])
            except sqlite3.IntegrityError:
                pass
       # self._database.commit()

    @property
    def num_entries(self):
        # cursor = self._database.execute("SELECT hash FROM traj_info;")
        # return len(cursor.fetchall())
        c = self._database.execute("SELECT COUNT(hash) from traj_info;").fetchone()
        return int(c[0])

    def set(self, traj_info):
        import sqlite3
        values = (
            traj_info.hash_value, traj_info.length, traj_info.ndim,
            np.array(traj_info.offsets), traj_info.abs_path, TrajectoryInfoCache.DB_VERSION,
            # lru db
            self._database_from_key(traj_info.hash_value)
        )
        statement = ("INSERT INTO traj_info (hash, length, ndim, offsets, abs_path, version, lru_db)"
                     "VALUES (?, ?, ?, ?, ?, ?, ?)", values)
        try:
            with self._database as c:
                c.execute(*statement)
        except sqlite3.IntegrityError as ie:
            logger.debug("insert failed: %s", ie, exc_info=True)
            return

        self._update_time_stamp(hash_value=traj_info.hash_value)

        if self.filename is not None:
            current_size = os.stat(self.filename).st_size
            if (self.num_entries >= config.traj_info_max_entries or
                    # current_size is in bytes, while traj_info_max_size is in MB
                    1.*current_size / 1024**2 >= config.traj_info_max_size):
                logger.info("Cleaning database because it has too much entries or is too large.\n"
                            "Entries: %s. Size: %.2fMB. Configured max_entires: %s. Max_size: %sMB"
                            % (self.num_entries, (current_size*1.0 / 1024**2),
                               config.traj_info_max_entries, config.traj_info_max_size))
                self._clean(n=self.clean_n_entries)

    def get(self, key):
        cursor = self._database.execute("SELECT * FROM traj_info WHERE hash=?", (key,))
        row = cursor.fetchone()
        if not row:
            raise KeyError()
        info = self._create_traj_info(row)
        self._update_time_stamp(key)
        return info

    def _database_from_key(self, key):
        """
        gets the database name for the given key. Should ensure a uniform spread
        of keys over the databases in order to minimize waiting times. Since the
        database has to be locked for updates and multiple processes want to write,
        each process has to wait until the lock has been released.

        By default the LRU databases will be stored in a sub directory "traj_info_usage"
        lying next to the main database.

        :param key: hash of the TrajInfo instance
        :return: str, database path
        """
        if not self.filename:
            return None

        from pyemma.util.files import mkdir_p
        hash_value_long = int(key, 16)
        # bin hash to one of either 10 different databases
        # TODO: make a configuration parameter out of this number
        db_name = str(hash_value_long)[-1] + '.db'
        directory = os.path.dirname(self.filename) + os.path.sep + 'traj_info_usage'
        mkdir_p(directory)
        return os.path.join(directory, db_name)

    def _update_time_stamp(self, hash_value):
        """ timestamps are being stored distributed over several lru databases.
        The timestamp is a time.time() snapshot (float), which are seconds since epoch."""
        db_name = self._database_from_key(hash_value)
        if not db_name:
            db_name=':memory:'

        def _update():
            import sqlite3
            try:
                with sqlite3.connect(db_name, timeout=self.lru_timeout) as conn:
                    """ last_read is a result of time.time()"""
                    conn.execute('CREATE TABLE IF NOT EXISTS usage '
                                 '(hash VARCHAR(32), last_read FLOAT)')
                    conn.commit()
                    cur = conn.execute('select * from usage where hash=?', (hash_value,))
                    row = cur.fetchone()
                    if not row:
                        conn.execute("insert into usage(hash, last_read) values(?, ?)", (hash_value, time.time()))
                    else:
                        conn.execute("update usage set last_read=? where hash=?", (time.time(), hash_value))
                    conn.commit()
            except sqlite3.OperationalError:
                # if there are many jobs to write to same database at same time, the timeout could be hit
                logger.debug('could not update LRU info for db %s', db_name)

        # this could lead to another (rare) race condition during cleaning...
        #import threading
        #threading.Thread(target=_update).start()
        _update()

    @staticmethod
    def _create_traj_info(row):
        # convert a database row to a TrajInfo object
        try:
            hash = row[0]
            length = row[1]
            ndim = row[2]
            offsets = row[3]
            assert isinstance(offsets, np.ndarray)
            abs_path = row[4]
            version = row[5]

            info = TrajInfo()
            info._version = version
            if version == 2:
                info._hash = hash
                info._ndim = ndim
                info._length = length
                info._offsets = offsets
                info._abs_path = abs_path
            else:
                raise ValueError("unknown version %s" % version)
            return info
        except Exception as ex:
            logger.exception(ex)
            raise UnknownDBFormatException(ex)

    @staticmethod
    def _format_tuple_for_sql(value):
        value = tuple(str(v) for v in value)
        return repr(value)[1:-2 if len(value) == 1 else -1]

    def _clean(self, n):
        """
        obtain n% oldest entries by looking into the usage databases. Then these entries
        are deleted first from the traj_info db and afterwards from the associated LRU dbs.

        :param n: delete n% entries in traj_info db [and associated LRU (usage) dbs].
        """
        # delete the n % oldest entries in the database
        import sqlite3
        num_delete = int(self.num_entries / 100.0 * n)
        logger.debug("removing %i entries from db" % num_delete)
        lru_dbs = self._database.execute("select hash, lru_db from traj_info").fetchall()
        lru_dbs.sort(key=itemgetter(1))
        hashs_by_db = {}
        age_by_hash = []
        for k, v in itertools.groupby(lru_dbs, key=itemgetter(1)):
            hashs_by_db[k] = list(x[0] for x in v)

        # debug: distribution
        len_by_db = {os.path.basename(db): len(hashs_by_db[db]) for db in hashs_by_db.keys()}
        logger.debug("distribution of lru: %s", str(len_by_db))
        ### end dbg

        # collect timestamps from databases
        for db in hashs_by_db.keys():
            with sqlite3.connect(db, timeout=self.lru_timeout) as conn:
                rows = conn.execute("select hash, last_read from usage").fetchall()
                for r in rows:
                    age_by_hash.append((r[0], float(r[1]), db))

        # sort by age
        age_by_hash.sort(key=itemgetter(1))
        if len(age_by_hash)>=2:
            assert[age_by_hash[-1] > age_by_hash[-2]]
        ids = map(itemgetter(0), age_by_hash[:num_delete])
        ids = tuple(map(str, ids))

        sql_compatible_ids = SqliteDB._format_tuple_for_sql(ids)

        with self._database as c:
            c.execute("DELETE FROM traj_info WHERE hash in (%s)" % sql_compatible_ids)

            # iterate over all LRU databases and delete those ids, we've just deleted from the main db.
            # Do this within the same execution block of the main database, because we do not want the entry to be deleted,
            # in case of a subsequent failure.
            age_by_hash.sort(key=itemgetter(2))
            for db, values in itertools.groupby(age_by_hash, key=itemgetter(2)):
                values = tuple(v[0] for v in values)
                with sqlite3.connect(db, timeout=self.lru_timeout) as conn:
                        stmnt = "DELETE FROM usage WHERE hash IN (%s)" \
                                % SqliteDB._format_tuple_for_sql(values)
                        curr = conn.execute(stmnt)
                        assert curr.rowcount == len(values), curr.rowcount



from celery import group

#failed = [16, 23, 35, 59, 71, 78, 83, 92, 95, 116, 128, 140, 164, 167]
from analysis import calc_cov_single_feat

if __name__ == '__main__':
    tasks = [calc_cov_single_feat.run.s(id_) for id_ in range(0, 12)]
    jobs = group(tasks)
    result = jobs.apply_async()



import os
from glob import glob
import pyemma
from celery import Celery

print('pyemma path:', pyemma.__path__)
import numpy as np
import sys

from pyemma._ext.sklearn.parameter_search import ParameterGrid
from pyemma.coordinates import source, featurizer
from pyemma.coordinates.data._base.transformer import StreamingTransformer
from pyemma.coordinates.estimation.covariance import Covariances

test_systems = [
    '1FME',
    '2F4K',
    '2JOF',
    '2WAV',
    'A3D',
    'CLN025',
    'GTT',
    'lambda',
    'NuG2',
    'PRB',
    'UVF',
    'NTL9',
]

features = (#'xyz',
            #'backbone',
            #'side_sidechain_torsions',
            #'dist_ca',
            #'dist_ca_d1',
            #'dist_ca_d2',
            #'dist_ca_expd',
            #'dist_ca_c_0.4',
            #'dist_ca_c_0.5',
            #'dist_ca_c_0.6',
            #'dist_ca_c_0.8',
            #'dist_ca_c_1.0',
            #'shrake_ruply',
            #'res_mindist',
#'res_mindist_d1',
#'res_mindist_d2',
#'res_mindist_expd',
'res_mindist_c_0.4',
'res_mindist_c_0.5',
'res_mindist_c_0.6',
'res_mindist_c_0.8',
'res_mindist_c_1.0',
            )
#RAW_TRAJS_PATH = '/home/marscher/NO_BACKUP/data/feature_sel'
#PRE_CALC_FEATURES_PATH = '/home/marscher/NO_BACKUP/data/feature_sel/cached_features/'
RAW_TRAJS_PATH = '/group/ag_cmb/simulation-data/DESRES-Science2011-FastProteinFolding/'
PRE_CALC_FEATURES_PATH = '/group/ag_cmb/marscher/feature_sel/'
output_path = '/group/ag_cmb/marscher/feature_sel/'

files = {
    name: {feature: sorted(glob(PRE_CALC_FEATURES_PATH + '/*{sys}*{feat}*.h5'.format(sys=name, feat='res_mindist')))
           for feature in features
           # omit cartesian coordinates here, because there are not preprocessed.
           }
    for name in test_systems}

#print (files)

def create_traj_top_pairs(test_system):
    # match trajectory files with topology

    if 'NTL9' not in test_system:
        patterns = RAW_TRAJS_PATH + '/*' + test_system + '*/*' + test_system + '*/*.dcd'
        print(patterns)
        top = glob(RAW_TRAJS_PATH + '/*' + test_system + '*/*' + test_system + '*/*.pdb')[0]
    else:
        patterns = RAW_TRAJS_PATH + '/' + test_system + '*/*.dcd'
        top = glob(RAW_TRAJS_PATH + '/' + test_system + '*/*.pdb')[0]
    trajs = sorted(glob(patterns))
    assert trajs

    import re
    # lambda has 4 independent runs, remove empty sub lists later on.
    files_for_features_independent_runs = [[], [], [], []]
    for f in trajs:
        m = re.match('.*{sys}-([0-9]).*'.format(sys=test_system), f)
        assert m
        ind_run = int(m.group(1))
        files_for_features_independent_runs[ind_run].append(f)
    files_for_features_independent_runs = list(filter(lambda x: x, (e for e in files_for_features_independent_runs)))

    assert top
    assert files_for_features_independent_runs

    trajs_with_top = {'trajs': files_for_features_independent_runs, 'top': top}
    return trajs_with_top


grid = ParameterGrid([{
'test_system': test_systems,
'feature': ['backbone'],
                       'lag': [500],  # dt=200ps, lag=100ns

                       # + [('backbone', 'dist_ca'),
                       #                   ('res_mindist', 'shrake_ruply'),
                       #                   ('cartesian', 'backbone')
                       #                   ]}]
                       }])

print('grid len:', len(grid))
for i, e in enumerate(grid):
    print(i, e)
# sys.exit(0)


class DistTrans(StreamingTransformer):
    def __init__(self, scheme='d1', c=None):
        super(DistTrans, self).__init__(10000)
        assert scheme in ('d1', 'd2', 'expd')
        self.scheme = scheme
        self.c = c

    def _transform_array(self, X):
        if self.c is None:
            if self.scheme == 'd1':
                return 1.0 / X
            elif self.scheme == 'd2':
                return 1.0 / np.sqrt(X)
            elif self.scheme == 'expd':
                return np.exp(-X)
        else:
            # apply cut-off
            mask = X >= self.c
            X[mask] = 0.0
            X[~mask] = 1.0
            return X

    def describe(self):
        pass

    def dimension(self):
        return super(DistTrans, self).dimension()


broker_url = 'amqp://marscher:foobar@wallaby:5672/myvhost'
app = Celery('tasks', broker=broker_url)


@app.task
def run(id):
    print('job number: ', id)
    params = grid[id]
    feature = params['feature']
    test_system = params['test_system']
    lag = params['lag']

    fname = output_path + 'scores_test_sys_{test_system}_lag_{lag}_feat_{feature}.npz'.format(
        test_system=test_system,
        lag=lag,
        mode='sliding',
        splitter='kfold',
        scoring_method='vamp2',
        feature=feature
    )
    print(fname)
    if os.path.exists(fname):
        print("results file %s already exists. Skipping" % fname)
        return

    try:
        files_for_features = [files[test_system][feature]]
        # lambda has 4 independent runs, remove empty sub lists later on.
        files_for_features_independent_runs = [[], [], [], []]
        import re
        for flist in files_for_features:
            for f in flist:
                m = re.match('.*{sys}-([0-9]).*'.format(sys=test_system), f)
                assert m
                ind_run = int(m.group(1))
                files_for_features_independent_runs[ind_run].append(f)
        # remove empty sub lists.
        files_for_features_independent_runs = list(
            filter(lambda x: x, (e for e in files_for_features_independent_runs)))
        assert files_for_features_independent_runs[0]
        #  create FragmentedReader per feature, which groups all the files of a system.
        # independent runs for each test system are named like $test_system-[0-9]-feat.dcd.h5
        reader = source(files_for_features_independent_runs, chunk_size=6000)

        # transform feature if needed (distances).
        if feature.startswith('res_mindist_'):
            name = feature
            match = re.match('res_mindist_c_(.*)', feature)
            if match is not None:
                arg = {'c': float(match.group(1))}
            else:
                arg = {'scheme': re.match('res_mindist_(.*)', name).group(1)}
            trans = DistTrans(**arg)
            trans.data_producer = reader
            reader = trans

        from sklearn.model_selection import KFold
        splitter_input = KFold(n_splits=10)
        cov = Covariances(n_covs=30, tau=lag, mode='sliding', n_save=4)

        cov.estimate(reader)
        scores = cov.score_cv(k=10, scoring_method='vamp2', splitter=splitter_input)
        parameters = np.array({'lag': lag,
                               'splitter': 'kfold',
                               'mode': 'sliding',
                               'test_system': test_system,
                               'scoring_method': 'vamp2',
                               })

        results = {'scores': scores, 'covs_': cov.covs_}

        np.savez_compressed(fname, parameters=parameters, **results)

    except BaseException as e:
        print('bad:', e, 'id:', id)
        import traceback
        traceback.print_last()

    print('successful')


if __name__ == '__main__':
    try:
        id_ = int(sys.argv[1])
        run(id_)
    except (IndexError, ValueError):
        print('provide an id as argument')
        sys.exit(1)




from tasks import transform_angles_cossin

from celery import group

if __name__ == '__main__':
    from glob import glob
    files = glob('/group/ag_cmb/marscher/feature_sel/*backbone*')
    tasks = [transform_angles_cossin.s(f) for f in files]
    print("num tasks:", len(tasks))
    jobs = group(tasks)
    result = jobs.apply_async()




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Created on 07.04.2015

@author: marscher
'''


import shutil
import tempfile
import unittest

from pyemma.coordinates.data.numpy_filereader import NumPyFileReader
from logging import getLogger


import numpy as np


class TestNumPyFileReader(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        cls.logger = getLogger('pyemma.'+cls.__class__.__name__)

        d = np.arange(3 * 100).reshape((100, 3))
        cls.d2 = np.arange(300, 900).reshape((200,3))
        d_1d = np.random.random(100)

        cls.dir = tempfile.mkdtemp(prefix='pyemma_npyreader')

        cls.f1 = tempfile.mktemp(suffix='.npy', dir=cls.dir)
        cls.f2 = tempfile.mktemp(suffix='.npy', dir=cls.dir)
        cls.f3 = tempfile.mktemp(suffix='.npz', dir=cls.dir)
        cls.f4 = tempfile.mktemp(suffix='.npy', dir=cls.dir)


        # 2d
        np.save(cls.f1, d)
        np.save(cls.f4, cls.d2)

        # 1d
        np.save(cls.f2, d_1d)

        np.savez(cls.f3, d, d)

        cls.files2d = [cls.f1, cls.f4] #cls.f3]
        cls.files1d = [cls.f2]
        cls.d = d
        cls.d_1d = d_1d

        cls.npy_files = [f for f in cls.files2d if f.endswith('.npy')]
        cls.npz = cls.f3

        return cls

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.dir, ignore_errors=True)

    def test_only_npy(self):
        reader = NumPyFileReader(self.npy_files)

        from_files = [np.load(f) for f in self.npy_files]
        concatenated = np.vstack(from_files)

        output = reader.get_output()

        self.assertEqual(reader.number_of_trajectories(), len(self.npy_files))
        self.assertEqual(reader.n_frames_total(), concatenated.shape[0])

        for x, y in zip(output, from_files):
            np.testing.assert_array_almost_equal(x, y)

    def test_small_chunks(self):
        reader = NumPyFileReader(self.npy_files)
        reader.chunksize = 30

        from_files = [np.load(f) for f in self.npy_files]
        concatenated = np.vstack(from_files)

        output = reader.get_output()

        self.assertEqual(reader.number_of_trajectories(), len(self.npy_files))
        self.assertEqual(reader.n_frames_total(), concatenated.shape[0])

        for x, y in zip(output, from_files):
            np.testing.assert_array_almost_equal(x, y)

    def test_skip(self):
        for skip in [0, 3, 13]:
            r1 = NumPyFileReader(self.npy_files[0])
            out_with_skip = r1.get_output(skip=skip)[0]
            r2 = NumPyFileReader(self.npy_files[0])
            out = r2.get_output()[0]
            np.testing.assert_almost_equal(out_with_skip, out[skip::],
                                           err_msg="The first %s rows were skipped, but that did not "
                                                   "match the rows with skip=0 and sliced by [%s::]" % (skip, skip))

    def test_skip_input_list(self):
        for skip in [0, 3, 13]:
            r1 = NumPyFileReader(self.npy_files)
            out_with_skip = r1.get_output(skip=skip)
            r2 = NumPyFileReader(self.npy_files)
            out = r2.get_output()
            for i in range(0, len(self.npy_files)):
                np.testing.assert_almost_equal(
                    out_with_skip[i], out[i][skip::],
                    err_msg="The first %s rows of the %s'th file were skipped, but that did not "
                            "match the rows with skip=0 and sliced by [%s::]" % (skip, i, skip)
                )

    def testSingleFile(self):
        reader = NumPyFileReader(self.npy_files[0])

        self.assertEqual(reader.n_frames_total(), self.d.shape[0])

    @unittest.skip("npz currently unsupported")
    def test_npz(self):
        reader = NumPyFileReader(self.npz)

        all_data = reader.get_output()

        fh = np.load(self.npz)
        data = [x[1] for x in list(fh.items())]
        fh.close()

        self.assertEqual(reader.number_of_trajectories(), len(data))

        for outp, inp in zip(all_data, data):
            np.testing.assert_equal(outp, inp)

    def test_stridden_access(self):
        reader = NumPyFileReader(self.f1)
        reader.chunksize = 10

        wanted = np.load(self.f1)

        for stride in [2, 3, 5, 7, 15]:
            first_traj = reader.get_output(stride=stride)[0]
            np.testing.assert_equal(first_traj, wanted[::stride],
                                    "did not match for stride %i" % stride)

    def test_lagged_stridden_access(self):
        reader = NumPyFileReader(self.f1)
        strides = [2, 3, 5, 7, 15]
        lags = [1, 3, 7, 10, 30]
        for stride in strides:
            for lag in lags:
                chunks = []
                for _, _, Y in reader.iterator(stride, lag):
                    chunks.append(Y)
                chunks = np.vstack(chunks)
                np.testing.assert_equal(chunks, self.d[lag::stride])

    def test_lagged_stridden_access_multiple_files(self):
        reader = NumPyFileReader(self.files2d)
        strides = [2, 3, 5, 7, 15]
        lags = [1, 3, 7, 10, 30]
        for stride in strides:
            for lag in lags:
                chunks = {i: [] for i in range(reader.number_of_trajectories())}
                for itraj, _, Y in reader.iterator(stride, lag):
                    chunks[itraj].append(Y)

                for i, k in enumerate(chunks.values()):
                    stack = np.vstack(k)
                    d = np.load(self.files2d[i])
                    np.testing.assert_equal(stack, d[lag::stride],
                                            "not equal for stride=%i"
                                            " and lag=%i" % (stride, lag))

    def test_usecols(self):
        reader = NumPyFileReader(self.f4)
        cols = (0, 2)
        it = reader.iterator(chunk=0, return_trajindex=False, cols=cols)
        with it:
            for x in it:
                np.testing.assert_equal(x, self.d2[:, cols])

    def test_different_shapes_value_error(self):
        with tempfile.NamedTemporaryFile(delete=False, suffix='.npy') as f:
            x = np.zeros((3, 42))
            np.save(f.name, x)
            myfiles = self.files2d[:]
            myfiles.insert(1, f.name)

            with self.assertRaises(ValueError) as cm:
                NumPyFileReader(myfiles)
            self.assertIn("different dimensions", cm.exception.args[0])

    def test_describe(self):
        r = NumPyFileReader(self.files2d)
        r.describe()

if __name__ == "__main__":
    unittest.main()
